// DrawAndDrawDoc.cpp : implementation of the CDrawAndDrawDoc class
//

#include "stdafx.h"
#include "DrawAndDraw.h"

#include "DrawAndDrawDoc.h"
#include "DrawViewSize.h"

#include "MainFrm.h"
#include "DrawAndDrawView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CDrawAndDrawDoc

IMPLEMENT_DYNCREATE(CDrawAndDrawDoc, CDocument)

BEGIN_MESSAGE_MAP(CDrawAndDrawDoc, CDocument)
	//{{AFX_MSG_MAP(CDrawAndDrawDoc)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CDrawAndDrawDoc construction/destruction

CDrawAndDrawDoc::CDrawAndDrawDoc()
{
    m_cx = -1;
    m_cy = -1;
	// TODO: add one-time construction code here

}

CDrawAndDrawDoc::~CDrawAndDrawDoc()
{
}

BOOL CDrawAndDrawDoc::OnNewDocument()
{
	if (!CDocument::OnNewDocument())
		return FALSE;

    if ((m_cx == -1) && (m_cy == -1)) {
        m_cx = 24;
        m_cy = 24;
    }
    else { // Prompt for grid size
        CDrawViewSize dlg;
        dlg.m_cx = m_cx;
        dlg.m_cy = m_cy;

        dlg.DoModal ();

        m_cx = dlg.m_cx;
        m_cy = dlg.m_cy;
    }

    int nByteCount = ((m_cx * m_cy) + 7) / 8;

    try { // Allocate grid memory
        m_byGrid.SetSize (nByteCount);
        m_byBuffer.SetSize (nByteCount);
    }
    catch (CMemoryException* e) { // Just in case...
        e->Delete ();
        return FALSE;
    }

    for (int i=0; i<nByteCount; i++) {
        m_byGrid[i] = 0;
        m_byBuffer[i] = 0;
    }
	// TODO: add reinitialization code here
	// (SDI documents will reuse this document)

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CDrawAndDrawDoc serialization

void CDrawAndDrawDoc::Serialize(CArchive& ar)
{	if (ar.IsStoring())
	{
		ar << (WORD) m_cx << (WORD) m_cy;
		// TODO: add storing code here
	}
	else
	{
		WORD cx, cy;
        ar >> cx >> cy;
        m_cx = (int) cx;
        m_cy = (int) cy;
		// TODO: add loading code here
	}
	m_byGrid.Serialize (ar);
    m_byBuffer.SetSize (m_byGrid.GetSize ());
}

/////////////////////////////////////////////////////////////////////////////
// CDrawAndDrawDoc diagnostics

#ifdef _DEBUG
void CDrawAndDrawDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CDrawAndDrawDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CDrawAndDrawDoc commands

void CDrawAndDrawDoc::DeleteContents() 
{ 
	m_byGrid.RemoveAll ();
    m_byBuffer.RemoveAll ();
	// TODO: Add your specialized code here and/or call the base class
	
	CDocument::DeleteContents();
}

CSize CDrawAndDrawDoc::GetGridSize ()
{
    return CSize (m_cx, m_cy);
}

BOOL CDrawAndDrawDoc::GetCell (int nRow, int nCol)
{
    if ((nRow < 0) || (nRow >= m_cy) ||
        (nCol < 0) || (nCol >= m_cx))
        return FALSE;

    int nBitNumber = ((nRow * m_cx) + nCol) % 8;
    int nByteNumber = ((nRow * m_cx) + nCol) / 8;

    return m_byGrid[nByteNumber] & (1 << nBitNumber);
}

BOOL CDrawAndDrawDoc::ToggleCell (int nRow, int nCol)
{
    if ((nRow < 0) || (nRow >= m_cy) ||
        (nCol < 0) || (nCol >= m_cx))
        return FALSE;

    int nBitNumber = ((nRow * m_cx) + nCol) % 8;
    int nByteNumber = ((nRow * m_cx) + nCol) / 8;

    m_byGrid[nByteNumber] ^= (1 << nBitNumber);
    SetModifiedFlag ();

    return m_byGrid[nByteNumber] & (1 << nBitNumber);
}

int CDrawAndDrawDoc::Evolve ()
{
    int nPrevRow, nPrevCol;
    int nNextRow, nNextCol;
    int nNeighbors, i, j;
    int nLiveCells = 0;
    BOOL bAlive;

    for (i=0; i<m_cy; i++) { // Do rows on the outside loop...
        if (m_bOpenEdge) {
            nPrevRow = (i == 0) ? m_cy - 1 : i - 1;
            nNextRow = (i == (m_cy - 1)) ? 0 : i + 1;
        }
        else {
            nPrevRow = (i == 0) ? -1 : i - 1;
            nNextRow = (i == (m_cy - 1)) ? -1 : i + 1;
        }

        for (j=0; j<m_cx; j++) { // ...and columns on the inside
            if (m_bOpenEdge) {
                nPrevCol = (j == 0) ? m_cx - 1 : j - 1;
                nNextCol = (j == (m_cx - 1)) ? 0 : j + 1;
            }
            else {
                nPrevCol = (j == 0) ? -1 : j - 1;
                nNextCol = (j == (m_cx - 1)) ? -1 : j + 1;
            }

            nNeighbors = 0;

            if (GetCell (nPrevRow, nPrevCol))
                nNeighbors++;
            if (GetCell (nPrevRow, j))
                nNeighbors++;
            if (GetCell (nPrevRow, nNextCol))
                nNeighbors++;

            if (GetCell (i, nPrevCol))
                nNeighbors++;
            if (GetCell (i, nNextCol))
                nNeighbors++;

            if (GetCell (nNextRow, nPrevCol))
                nNeighbors++;
            if (GetCell (nNextRow, j))
                nNeighbors++;
            if (GetCell (nNextRow, nNextCol))
                nNeighbors++;

            bAlive = GetCell (i, j);

            if ((bAlive && ((nNeighbors == 2) ||
                (nNeighbors == 3))) || (!bAlive && (nNeighbors == 3))) {
                SetBackgroundCell (i, j, TRUE);
                nLiveCells++;
            }
            else
                SetBackgroundCell (i, j, FALSE);
        }
    }

    int nByteCount = ((m_cx * m_cy) + 7) / 8;
    for (i=0; i<nByteCount; i++)
        m_byGrid[i] = m_byBuffer[i];

    SetModifiedFlag ();
    UpdateAllViews (NULL, 1);
    return nLiveCells;
}

void CDrawAndDrawDoc::SetBackgroundCell (int nRow, int nCol, BOOL bState)
{
    int nBitNumber = ((nRow * m_cx) + nCol) % 8;
    int nByteNumber = ((nRow * m_cx) + nCol) / 8;

    if (bState)
        m_byBuffer[nByteNumber] |= (1 << nBitNumber);
    else
        m_byBuffer[nByteNumber] &= ~(1 << nBitNumber);
}
